__all__ = [
    "DataJoinGraph"
]

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Dict, Set, Union, Optional
from graphviz import Digraph

from ..utils.autoai.errors import OBMMainNodeAlreadySet
from ..utils.autoai.enums import TShirtSize, PredictionType
from ..utils.autoai.utils import is_ipython, check_graphviz_binaries


class BaseOBMJson:
    """Base class for helper objects representation."""

    def __repr__(self):
        return self.__str__()


class Node(BaseOBMJson):
    """Node class for json representation and conversion of graphviz nodes."""

    def __init__(self, name: str, timestamp_column_name: str = None, timestamp_format: str = None,
                 csv_separator: str = None, sheet_name: str = None):
        self.table = Table(name=name, timestamp_column_name=timestamp_column_name, timestamp_format=timestamp_format,
                           csv_separator=csv_separator, sheet_name=sheet_name)

    def to_dict(self):
        """Convert this Node to dictionary for further REST API call."""
        return {"table_name": self.table.name}

    def _pretty_print(self, main: bool = False) -> str:
        if main:
            return f"data_join_graph.node(name='{self.table.name}', main=True)\n"

        else:
            return f"data_join_graph.node(name='{self.table.name}')\n"

    def __str__(self):
        return (f"\n\t\tNODE:\n"
                f"\t\t\t{self.table.__str__()}")


class Edge(BaseOBMJson):
    """Edge helper class for json representation of graphviz edge."""

    def __init__(self, from_node: str, to_node: str, from_column: List[str], to_column: List[str]):
        if not isinstance(from_column, list) or not isinstance(to_column, list):
            raise TypeError("\"from_column\" and \"to_column\" need to be of type List[str].")

        self.from_node = from_node
        self.to_node = to_node
        self.from_column = from_column
        self.to_column = to_column

    def to_dict(self):
        """Convert this Node to dictionary for further REST API call."""
        _dict = {
            "from": self.from_node,
            "to": self.to_node,
            "from_column": self.from_column,
            "to_column": self.to_column,
        }
        return _dict

    def _pretty_print(self) -> str:
        return f"data_join_graph.edge(from_node='{self.from_node}', to_node='{self.to_node}', from_column={self.from_column}, to_column={self.to_column})\n"

    def __str__(self):
        return (f"\n\t\tEDGE:\n"
                f"\t\t\tFROM: {self.from_node}\n"
                f"\t\t\tTO: {self.to_node}\n"
                f"\t\t\tFROM COLUMN: {self.from_column}\n"
                f"\t\t\tTO COLUMN: {self.to_column}")


class Table(BaseOBMJson):
    """Table class to represent / define OBM tables."""

    def __init__(self, name: str, timestamp_column_name: str = None, timestamp_format: str = None,
                 csv_separator: str = None, sheet_name: str = None):
        if (timestamp_column_name or timestamp_format) and not (timestamp_column_name and timestamp_format):
            print("Need to pass both column name and date format in order to mark column as timestamp type.")

        self.name = name
        self.timestamp_column_name = timestamp_column_name
        self.timestamp_format = timestamp_format
        self.csv_separator = csv_separator
        self.sheet_name = sheet_name
        self.source = {}

    def to_dict(self):
        """Convert this Node to dictionary for further REST API call."""
        _dict = {
            "table_source": self.source
        }
        if self.timestamp_column_name and self.timestamp_format:
            _dict.update({
                "column_format": {
                    self.timestamp_column_name: self.timestamp_format,
                },
                "timestamp_column_name": self.timestamp_column_name
            })
        if self.csv_separator:
            _dict.update({
                "field_delimiter": self.csv_separator
            })

        if self.sheet_name:
            _dict.update({
                "sheet_name": self.sheet_name
            })

        return _dict

    def __str__(self):
        return (f"TABLE:\n"
                f"\t\t\t\tNAME: {self.name}\n"
                f"\t\t\t\tSOURCE: {self.source}\n" +
                (
                        f"\t\t\t\tTIMESTAMP_COLUMN_NAME: \'{self.timestamp_column_name}\'\n"
                        f"\t\t\t\tCOLUMN_FORMAT: {{\n\t\t\t\t\t'{self.timestamp_column_name}': '{self.timestamp_format}'\n"
                        f"\t\t\t\t}}" if self.timestamp_column_name and self.timestamp_format else ""
                ) +
                (f"\t\t\t\tFIELD_DELIMITER: '{self.csv_separator}'\n" if self.csv_separator else "") +
                (f"\t\t\t\tSHEET_NAME: '{self.sheet_name}'\n" if self.sheet_name else "")
        )


class DataJoinGraph(Digraph, BaseOBMJson):
    """
    DataJoinGraph class - helper class for handling multiple data sources for AutoAI experiment.

    You can define the overall relations between each of data source and see these defined relations
    in a form of string representation calling print(ObmGraph) or to leverage graphviz library
    and make entire graph visualization.

    Parameters
    ----------
    t_shirt_size: enum TShirtSize, optional
        The size of the computation POD.

    Example
    -------
    >>> data_join_graph = DataJoinGraph()
    >>> # or
    >>> data_join_graph = DataJoinGraph(t_shirt_size=DataJoinGraph.TShirtSize.L)
​
    >>> data_join_graph.node(name="main")
    >>> data_join_graph.node(name="customers")
    >>> data_join_graph.node(name="transactions")
    >>> data_join_graph.node(name="purchases")
    >>> data_join_graph.node(name="products")
​
    >>> data_join_graph.edge(from_node="main", to_node="customers",
    >>>                from_column=["group_customer_id"], to_column=["group_customer_id"])
    >>> data_join_graph.edge(from_node="main", to_node="transactions",
    >>>                from_column=["transaction_id"], to_column=["transaction_id"])
    >>> data_join_graph.edge(from_node="main", to_node="purchases",
    >>>                from_column=["group_id"], to_column=["group_id"])
    >>> data_join_graph.edge(from_node="transactions", to_node="products",
    >>>                from_column=["product_id"], to_column=["product_id"])
​
    >>> print(data_join_graph)
    >>> data_join_graph.visualize()
    """
    TShirtSize = TShirtSize
    NodeTemplate = """<
    <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="1">
      <TR>
          <TD STYLE="ROUNDED" BGCOLOR=\"{color}\" COLSPAN=\"{colspan}\">{dataset_name}</TD>
      </TR>
      <TR>
      </TR>
    </TABLE>>"""
    ColumnsIndex = NodeTemplate.rfind("<TR>") + len("<TR>")
    ColumnTemplate = "<TD STYLE=\"ROUNDED\" PORT=\"{port}\">{column_name}</TD>"

    def __init__(self,
                 t_shirt_size: 'TShirtSize' = TShirtSize.M,
                 max_depth: int = 3,
                 data_source_type: str = "csv",
                 worker_nodes_number: int = 1):
        super().__init__(comment='Data Join Graph')

        self.t_shirt_size = t_shirt_size
        self.worker_nodes_number = worker_nodes_number

        # note: these values should be provided by optimizer
        self.target_column = None
        self.__problem_type = None
        # --- end note

        self.main_node_name = None  # main node by default the first one
        self.main_set = False
        self.max_depth = max_depth
        self.data_source_type = data_source_type

        self.nodes: List['Node'] = []
        self.edges: List['Edge'] = []
        self.edges_to_ports: Dict[str, str] = {}
        self.node_edges: Dict[str, Set[str]] = {}

        self.fairness_info = None

    @property
    def problem_type(self) -> 'PredictionType':
        return self.__problem_type

    @problem_type.setter
    def problem_type(self, value: 'PredictionType') -> None:
        """We need to map prediction types between KB and OBM."""
        if value == PredictionType.MULTICLASS:
            self.__problem_type = 'classification'
        else:
            self.__problem_type = value

    def node(self,
             name: str,
             timestamp_column_name: Optional[str] = None,
             timestamp_format: Optional[str] = None,
             csv_separator: Optional[str] = None,
             sheet_name: Optional[str] = None,
             main: Optional[bool] = False) -> None:
        """
        Add node to the graph. The node is representing the particular data source for training.
        The node added as a first one will be set as main one. If you want to set another node as a main, please
        set a `main` argument to True.

        Parameters
        ----------
        name: str, required
            The name/id of the node, it must be the same as id passed to particular linked DataConnection.
        timestamp_column_name: str, optional
            The name of the column, which includes dates.
        timestamp_format: str, optional
            Format of dates contained in column named 'timestamp_column_name'.
        csv_separator: str, optional
            Separator used in source csv file.
        sheet_name: str, optional
            Used for xlsx files to indicate which sheet is used.
        main: bool, optional
            If you want to set this node as a main one, please set a `main` argument to True.
        Example
        -------
        >>> data_join_graph.node(name="main")
        """
        self.node_edges[name] = set()

        # note: if we do not have any nodes already added, first one will be the main node
        if not self.nodes:
            self.main_node_name = name

        # override the default main node
        if main and not self.main_set:
            self.main_node_name = name
            self.main_set = True

        elif main:
            raise OBMMainNodeAlreadySet(self.main_node_name)

        self.nodes.append(Node(name=name, timestamp_column_name=timestamp_column_name,
                               timestamp_format=timestamp_format, csv_separator=csv_separator, sheet_name=sheet_name))
        # --- end note

    def edge(self, from_node, to_node, from_column: List[str], to_column: List[str]):
        """
        Add edge to the graph. The edge defines the connection between two DataConnections.
        eg. main --- from column customer_id to column customer_id --> customers

        Parameters
        ----------
        from_node: str, required
            The starting Node.

        to_node: str, required
            The ending Node

        from_column: List[str], required
            The list of columns located in the starting Node in order, these columns will be connected to
            the ending Node columns.

        to_column: List[str], required
            The list of columns located in the ending Node.

        Example
        -------
        >>> data_join_graph.edge(from_node="main", to_node="transactions",
        >>>                from_column=["transaction_id"], to_column=["transaction_id"])
        """
        self.node_edges[from_node] |= set(from_column)
        self.node_edges[to_node] |= set(to_column)

        self.edges.append(Edge(from_node=from_node, to_node=to_node, from_column=from_column, to_column=to_column))

    def _visualize_nodes(self):
        for node in self.nodes:
            name = node.table.name
            column_names = self.node_edges[name]
            self.edges_to_ports.update({f"{name}{column}": f"{name}:e{i}" for i, column in enumerate(column_names)})
            columns = ''.join(
                [self.ColumnTemplate.format(port=f"e{i}", column_name=name) for i, name in enumerate(column_names)])
            node_label = self.NodeTemplate[:self.ColumnsIndex] + columns + self.NodeTemplate[self.ColumnsIndex:]
            color = "LIGHTBLUE2" if self.main_node_name == name else "WHITE"

            super().node(name=name, label=node_label.format(color=color,
                                                            dataset_name=name,
                                                            colspan=str(len(column_names))))

    def _visualize_edges(self):
        for edge in self.edges:
            edges = [(self.edges_to_ports[edge.from_node + from_col], self.edges_to_ports[edge.to_node + to_col])
                     for from_col, to_col in zip(edge.from_column, edge.to_column)]

            super().edges(edges)

    @check_graphviz_binaries
    def visualize(self):
        super().clear()
        super().attr('node', shape='plaintext')
        super().attr('edge', minlen='2')
        super().attr('graph', nodesep='1')

        self._visualize_nodes()
        self._visualize_edges()

        """Display graph in the notebook or as a rendered image."""
        if is_ipython():
            import IPython.display
            IPython.display.display(self)

        else:
            self.render(view=True)

    def to_dict(self):
        """Convert this Node to dictionary for further REST API call."""
        _dict = {
            "id": "obm",
            "type": "execution_node",
            "op": "kube",
            "runtime_ref": "obm",
            "inputs": [{"id": node.table.name} for node in self.nodes],
            "outputs": [
                {
                    "id": "obm_out"
                }
            ],
            "parameters": {
                "stage_flag": True,
                "output_logs": True,
                "engine": {
                    "template_id": "spark-3.2-automl-template",
                },
                "obm": {
                    "Entity_Graph": {
                        "nodes": [node.to_dict() for node in self.nodes],
                        "edges": [edge.to_dict() for edge in self.edges]
                    },
                    "Tables": {node.table.name: node.table.to_dict() for node in self.nodes},
                    "Feature_Selector": {
                        "selectors": [
                            "deduplicate",
                            "consistent"
                        ]
                    },
                    "OneButtonMachine": {
                        "main_table": self.main_node_name,
                        "target_column": self.target_column,
                        "max_depth": self.max_depth,
                        "data_source": self.data_source_type,
                        "problem_type": self.problem_type,
                        "join_limit": 50,
                    }
                }
            }
        }

        if self.fairness_info is not None:
            _dict['parameters']['fairness_info'] = self.fairness_info

        return _dict

    @classmethod
    def _from_dict(cls, _dict: dict) -> 'DataJoinGraph':
        """Create data join graph object from wml pipeline parameters."""
        data_join_graph = cls()

        for node in _dict['parameters']['obm']['Entity_Graph']['nodes']:
            if node['table_name'] == _dict['parameters']['obm']['OneButtonMachine']['main_table']:
                data_join_graph.node(name=node['table_name'], main=True)

            else:
                data_join_graph.node(name=node['table_name'])

        [data_join_graph.edge(
            from_node=edge['from'],
            to_node=edge['to'],
            from_column=edge['from_column'],
            to_column=edge['to_column']) for edge in _dict['parameters']['obm']['Entity_Graph']['edges']]

        data_join_graph.fairness_info = _dict['parameters'].get('fairness_info')

        return data_join_graph

    def pretty_print(self, show_imports: bool = True, ipython_display: bool = False) -> Union[str, None]:
        """Returns the Python source code representation of the DataJoinGraph.

        Parameters
        ----------
        show_imports : bool, default True

            Whether to include import statements in the pretty-printed code.

        ipython_display : union type, default False

            - False

              Return the pretty-printed code as a plain old Python string.

            - True:

              Pretty-print in notebook cell output with syntax highlighting.

            - 'input'

              Create a new notebook cell with pretty-printed code as input.

        Returns
        -------
        str or None
            If called with ipython_display=False, return pretty-printed Python source code as a Python string.
        """
        nodes_str = ""
        for node in self.nodes:
            if self.main_node_name == node.table.name:
                nodes_str = nodes_str + node._pretty_print(main=True)

            else:
                nodes_str = nodes_str + node._pretty_print()

        edges_str = ""
        for edge in self.edges:
            edges_str = edges_str + edge._pretty_print()

        result = f"""{'from ibm_watson_machine_learning.preprocessing import DataJoinGraph' if show_imports else ''}
data_join_graph = DataJoinGraph()
{nodes_str}
{edges_str}
"""

        if ipython_display == False:
            return result

        elif ipython_display == 'input':
            import IPython.core
            ipython = IPython.core.getipython.get_ipython()
            comment = "# generated by pretty_print(ipython_display='input') from previous cell\n"
            ipython.set_next_input(comment + result, replace=False)

        else:
            assert ipython_display in [True, 'output']
            import IPython.display
            markdown = IPython.display.Markdown(f'```python\n{result}\n```')
            return IPython.display.display(markdown)

    def __str__(self):
        return (f"\nDataJoinGraph:\n"
                f"\tMAIN NODE: {self.main_node_name}\n"
                f"\tTARGET COLUMN: {self.target_column}\n"
                f"\tMAX DEPTH: {self.max_depth}\n"
                f"\tDATA SOURCE TYPE: {self.data_source_type}\n"
                f"\tPROBLEM TYPE: {self.problem_type}\n"
                f"\tNODES: {[node for node in self.nodes]}\n"
                f"\tEDGES: {[edge for edge in self.edges]}"
                )



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import os

from .globalization_util import GlobalizationUtil
import ibm_watson_machine_learning.messages


def get_message_dict(locale):
    file_name = "messages_" + locale + ".json"
    message_dict = {}
    path = os.path.dirname(ibm_watson_machine_learning.messages.__file__)
    messages = []
    # try to load the respective json file for the locale
    try:
        with open(os.path.join(path, file_name)) as f:
            messages = json.loads(f.read())
    # load the english dictionary if the json file for the locale doesn't exist
    except:
        try:
            return get_message_dict("en")
        except:
            raise Exception(
                "An error occurred while trying to load the message json file for the {} locale. "
                "Make sure the json file exists and is located in the correct folder.".format(locale))

    messages_list = [{item["code"]:item["message"]} for item in messages]
    for i in range(len(messages_list)):
        message_dict.update(messages_list[i])
    return message_dict


MESSAGE_DICT = get_message_dict(GlobalizationUtil.get_language())


def replace_args_in_message(message, *args):
    if args:
        varss = []
        for x in args:
            if x is not None and type(x) is not Exception:
                varss.append(x)
        if '{0}' in message:
            message = message.format(*varss)
        else:
            message = message % tuple(varss)
    return message


class Messages:

    @classmethod
    def get_message(cls, *args, message_id):
        message = MESSAGE_DICT.get(message_id)

        if args and message:
            message = replace_args_in_message(message, *args)

        return message



__all__ = [
    'is_run_id_exists'
]

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, Optional

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure


def is_run_id_exists(wml_credentials: Dict, run_id: str, space_id: Optional[str] = None) -> bool:
    """
    Check if specified run_id exists for WML client initialized with passed credentials.

    Parameters
    ----------
    wml_credentials: dictionary, required

    run_id: str, required
        Training run id of AutoAI experiment.

    space_id: str, optional
        Optional space id for WMLS and CP4D.
    """
    client = APIClient(wml_credentials)

    if space_id is not None:
        client.set.default_space(space_id)

    try:
        client.training.get_details(run_id, _internal=True)

    except ApiRequestFailure as e:
        if 'Status code: 404' in str(e):
            return False

        else:
            raise e

    return True



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import inspect
import logging
import operator
import itertools
import numpy as np
from functools import partial
from abc import ABC, abstractmethod
from timeit import default_timer as time

from sklearn.metrics import check_scoring
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OrdinalEncoder, LabelEncoder
from sklearn.ensemble._hist_gradient_boosting.loss import _LOSSES
from sklearn.utils.multiclass import check_classification_targets
from sklearn.ensemble._hist_gradient_boosting.common import Y_DTYPE
from sklearn.utils import check_X_y, check_random_state, check_array
from sklearn.ensemble._hist_gradient_boosting.binning import _BinMapper

from ibmfl.model.xgb_fl_model import XGBFLModel
from ibmfl.util.xgboost.utils import is_classifier
from ibmfl.util.xgboost.hyperparams import init_parameters

from ibmfl.exceptions import HyperparamsException
from ibmfl.exceptions import LocalTrainingException, ModelUpdateException
from ibmfl.party.training.local_training_handler import LocalTrainingHandler

logger = logging.getLogger(__name__)


class XGBoostBaseLocalTrainingHandler(LocalTrainingHandler, ABC):
    """
    Class implementation for XGBoost Base Local Training Handler
    """

    def update_model(self, model_update):
        """
        Update local model with model updates received from FusionHandler

        :param model_update: ModelUpdate
        :type model_update: `ModelUpdate`
        :return: `None`
        """
        try:
            # Check to see if worker's core model object has been initialized.
            if self.fl_model is None:
                self.fl_model = XGBFLModel('XGBFLModel', self.hyperparams)
                logger.info(
                    'No model update was provided, initialized new FL model object.')

            # Update Model Object
            if model_update is not None:
                self.fl_model.update_model(model_update)
                logger.info('Local model updated.')
            else:
                logger.info('No model update was provided.')
        except Exception as ex:
            raise LocalTrainingException(
                'No query information is provided. '+ str(ex))

    def train(self, fit_params=None):
        """
        Primary wrapper function used for routing internal remote function calls
        within the Local Training Handler functions.

        :param fit_params: A dictionary payload structure containing two key \
        signatures, `func` and `args`, which respectively are the target \
        function defined within the Local Training Handler and the arguments \
        defined within the executing function, which is defined as a dictionary \
        containing key-value pairs of matching arguments.
        :type fit_params: `dict`
        :return: Returns the corresponding values depending on the function \
        remotely called from the aggregator.
        """
        result = None
        try:
            # Validate Incoming Payload Parameter
            if fit_params is None:
                raise LocalTrainingException('Provided fit_params is None, no '
                                             'functions were executed.')

            # Validate Payload Signature
            if 'func' not in fit_params or 'args' not in fit_params:
                raise LocalTrainingException('Malformed payload, must include '
                                             'func and args in payload.')

            # Validate Defined Function Header
            if not (isinstance(fit_params['func'], str) and
                    hasattr(self, fit_params['func'])):
                raise LocalTrainingHandler('Function header is not valid or is '
                                           'not defined within the scope of the '
                                           'local training handler.')

            # Validate Payload Argument Parameter Mappings Against Function
            spec = inspect.getargspec(eval('self.'+fit_params['func']))
            for k in fit_params['args'].keys():
                if k not in spec.args:
                    raise LocalTrainingHandler('Specified parameter argument is '
                                               'not defined in the function.')

            # Construct Function Call Command
            result = eval('self.'+fit_params['func'])(**fit_params['args'])
        except Exception as ex:
            raise LocalTrainingException('Error processing remote function ' +
                                         'call: ' + str(ex))

        return result

    def initialize(self, params):
        """
        A remote function call which performs the following procedures:
        1. Obtains global hyperparameters from the aggregator and initializes
        them for each of the local worker's parameters.
        2. Performs the following set of preliminary validation checks
        and processes:
           a. Perform dataset encoding and validation checks.
           b. Set local worker's seed and PRNG states.
           c. Initialize key parameters from defined hyperparameters in 1.
           d. Check for missing data within each local worker's dataset.
           e. Returns various parameters needed for aggregator to perform
           tree growth operation.

        :param params: A hyperparameter dictionary from the aggregator.
        :type params: `dict`
        :return: Dictionary containing parameters necessary for tree growth at \
        the aggregator.
        :rtype: `dict`
        """
        # Initialize Hyperparameters
        logger.info('Recieved and initializing local hyperparameters.')
        self.hyperparams = params

        init_parameters(self, params)

        logger.info(
            'Performing preliminary checks, validation, and initialization.')

        # Data Validation and Encoding Checks
        logger.info('[CHECK] Dataset Encoding and Validation Checks')
        self.data_val_enc()

        # Check PRNG State
        logger.info('[CHECK] PRNG State and Seed Validation')
        self.check_prng_state()

        # Initialize Key Parameters
        logger.info('[CHECK] Initialize Key Parameters')
        self.loss_ = self.get_loss(sample_weight=self.sample_weight)

        # Validate Dataset
        self.X_train, self.y_train = self.X, self.Y
        self.X_val, self.y_val = None, None

        # Check for Missing Data
        logger.info('[CHECK] Check for Missing Data')
        self.has_missing_values = np.isnan(
            self.X_train).any(axis=0).astype(np.uint8)

        # Prepare Payload
        params = {
            'has_missing_values': self.has_missing_values,
            'need_update_leaves_values': self.loss_.need_update_leaves_values,
            'n_features_': self.X.shape[1]
        }
        if is_classifier(self):
            params['classes_'] = self.classes_

        return params

    def check_prng_state(self):
        """
        Given the initialize `random_state` from the hyperparameter, we set the
        random seed value of the local training handler.

        :return: `None`
        """
        # Check PRNG Random State
        rng = check_random_state(self.random_state)
        self._random_seed = rng.randint(np.iinfo(np.uint32).max, dtype='u8')

    def get_sample_size(self):
        """
        Obtain the number of samples from the training set.

        :return: Returns the number of samples in each worker.
        :rtype: `int`
        """
        if hasattr(self, 'X_train') and self.X_train is not None:
            return len(self.X_train)
        else:
            return 0

    def set_eps(self, eps):
        """
        Set the eps defined by the aggregator, which also defines the `max_bins`
        size in the local training handler for the histogram.

        :param eps: Epsilon parameter which dictates the bin size of the \
        histogram.
        :type eps: `float`
        :return: `None`
        """
        if isinstance(eps, float) and eps > 0:
            self.eps = eps
            self.max_bins = int(1 // self.eps)

            logger.info('Setting local worker eps to ' + str(self.eps))
            logger.info('Setting local worker max_bin to ' +
                        str(self.max_bins))
        else:
            raise ValueError('Defined eps is not valid.')

    def bin_data(self):
        """
        Initialize process to bin the training data. The number of bins
        allocated here are based on the maximum number of bins set by the
        aggregator and 1 more to account for the missing values.

        :return: `None`
        """
        logger.info('Binning Dataset')
        self.n_bins = self.max_bins + 1  # +1 for Missing Data
        self.bin_mapper_ = _BinMapper(
            n_bins=self.n_bins, random_state=self._random_seed)

        self.X_binned_train = self.bin_mapper_.fit_transform(self.X_train)
        self.n_samples = self.X_binned_train.shape[0]

    def init_null_preds(self):
        """
        Initialize null predictions from baseline models. These values are used
        to construct the initial tree structure of the model. Note that
        raw_predictions has shape (n_trees_per_iteration, n_samples) whereas
        n_trees_per_iteration is n_classes in multiclass classification, else 1.
        Furthermore, setups initial data structures for metrics and performs
        early stop validation checks.

        :return: `None`
        """
        # Generate Baseline Predictions
        logger.info('Generating Initial Predictions from Null Model')
        self.baseline = self.loss_.get_baseline_prediction(
            self.y_train, self.sample_weight, self.n_trees)
        self.raw_predictions = np.zeros(shape=(self.n_trees, self.n_samples),
                                        dtype=self.baseline.dtype) + self.baseline

        # Initialize Relevant Datastructures for Metrics
        self._predictors = predictors = []
        self._scorer = None
        self.raw_predictions_val = None
        self.train_score_ = []
        self.validation_score_ = []

        self.begin_at_stage = 0

        # Initialize Model Attributes
        self.fl_model.n_trees = self.n_trees
        self.fl_model._baseline_prediction = self.baseline
        self.fl_model._raw_predictions = self.raw_predictions
        self.fl_model.loss_ = self.loss_
        self.fl_model.n_features_ = self.X.shape[1]
        if is_classifier(self):
            self.fl_model.classes_ = self.classes_

    def init_grad(self):
        """
        Process to initialize gradients and hessians on the Local Training
        Handler. This is an empty data structure used to contain and later
        will be used to send over to aggregator.
        Note: shape = (n_trees_per_iteration, n_samples)

        :return: `None`
        """
        self.gradients, self.hessians = self.loss_.init_gradients_and_hessians(
            n_samples=self.n_samples,
            sample_weight=self.sample_weight,
            prediction_dim=self.n_trees
        )

    def update_grad(self):
        """
        A wrapper function to update the corresponding gradient and hessian
        statistics inplace based on the raw predictions from the previous
        iteration of the model.

        :return: `None`
        """
        self.loss_.update_gradients_and_hessians(self.gradients, self.hessians,
                                                 self.y_train,
                                                 self.raw_predictions,
                                                 self.sample_weight)

    def collect_hist(self, k):
        """
        We return values of the surrogate histogram values, gradients, hessians,
        and the number of non-missing bin counts from the local worker.

        :param k: The correseponding feature index of the dataset.
        :type k: `int`
        :return: Returns a tuple comprising of the histogram bin index values, \
        gradient, hessian, missing value counts per feature, bin threshold value.
        :rtype: (`np.array`, `np.array`, `np.array`, `list`, `list`)
        """
        return self.X_binned_train, self.gradients[k, :], self.hessians[k, :], \
            self.bin_mapper_.n_bins_non_missing_, self.bin_mapper_.bin_thresholds_

    def update_raw_preds(self, k, predictor, use_val):
        """
        Function which updates internal local trainer handler's raw_prediction
        values given the intermediate state of the model.

        NOTE: Because we merged the histogram, the histogram mappings no longer
        apply in this context, so all predictions must be done entirely from
        scratch again - cannot use the histogram subtraction technique from
        previous implementation here.

        :param k: The correseponding feature index of the dataset.
        :type k: `int`
        :param predictor: The predictor object state of the XGBoost tree.
        :type  predictor: `TreePredictor`
        :param use_val: Boolean indicator whether to use the validation set. \
        (Currently this parameter is not supported at the moment.)
        :type use_val: `Boolean`
        :return: `None`
        """
        self.raw_predictions[k] += predictor.predict(self.X_train)

    def data_val_enc(self):
        """
        Performs a validation procedure of the input dataset to correspondingly
        check the data types as well as the data encoding.

        :return: `None`
        """
        logging.info('Performing local worker data encoding and validation.')
        (x, y), (_) = self.data_handler.get_data()

        # Perform Data Encoding Validation
        self.X, y = check_X_y(x, y, dtype=[np.float64], force_all_finite=False)
        self.Y = self.encode_target(y)

        # TODO load sample_weight from config
        self.sample_weight = None

    @abstractmethod
    def encode_target(self, y=None):
        raise NotImplementedError


class XGBRegressorLocalTrainingHandler(XGBoostBaseLocalTrainingHandler):
    _VALID_LOSSES = ('least_squares')

    def encode_target(self, y):
        """
        Converts the input y to the expected dtype.

        :param y: The corresponding target data from the dataset to encode.
        :type y: `np.array`
        :return: Returns the corresponding encoded y values.
        :rtype: `np.array`
        """
        self.n_trees = 1
        return y.astype(Y_DTYPE, copy=False)

    def get_loss(self, sample_weight):
        """
        Given the initialized loss type defined under the hyerparameters, we
        return the corresponding loss function to dictate the corresponding
        learning task of the model.

        :param sample_weight: Weights of training data
        :type sample_weight: `np.ndarray`
        :return: Returns the respective loss object as defined in the FL \
        hyperparameters.
        :rtype: Derivation of `BaseLoss`
        """
        return _LOSSES[self.loss](sample_weight=sample_weight)


class XGBClassifierLocalTrainingHandler(XGBoostBaseLocalTrainingHandler):
    _VALID_LOSSES = ('binary_crossentropy', 'categorical_crossentropy', 'auto')

    def encode_target(self, y):
        """
        Converts the input y to the expected dtype and performs a label
        encoding. Here, we assume that each party has at least one sample of
        the corresponding class label type for each different classes.

        :param y: The corresponding target data from the dataset to encode.
        :type y: `np.array`
        :return: Returns the corresponding encoded y values.
        :rtype: `np.array`
        """
        # Validate Classification Target Values
        check_classification_targets(y)

        # Apply Label Encoder Transformation
        lab_enc = LabelEncoder()
        enc_y = lab_enc.fit_transform(y).astype(np.float64, copy=False)

        # Extract Encoded Target Sizes
        self.classes_ = lab_enc.classes_
        if self.classes_.shape[0] != self.num_classes:
            raise ValueError('Number of classes defined in configuration file '
                             'and the classes derived from the data does not '
                             'match. Found {0} classes, while config file '
                             'is defined as {1} classes.'.format(
                             self.classes_.shape[0], self.num_classes))

        if self.loss == 'auto':
            self.n_trees = 1 if self.classes_.shape[0] <= 2 else self.classes_.shape[0]
        else:
            self.n_trees = 1 if self.num_classes <= 2 else self.num_classes

        return enc_y

    def get_loss(self, sample_weight):
        """
        Given the initialized loss type defined under the hyerparameters, we
        return the corresponding loss function to dictate the corresponding
        learning task of the model. If auto is selected, then we will
        automatically determine whether the classification task is binary or
        multiclass given the label encoding cardinality.

        :param sample_weight: Weights of training data
        :type sample_weight: `np.ndarray`
        :return: Returns the respective loss object as defined in the FL \
        hyperparameters.
        :rtype: Derivation of `BaseLoss`
        """
        if (self.loss == 'categorical_crossentropy' and self.n_trees == 1):
            raise ValueError("Incompatible loss and target variable counts.")

        if self.loss == 'auto':
            return _LOSSES['binary_crossentropy'](sample_weight=sample_weight) \
                if self.n_trees == 1 else \
                _LOSSES['categorical_crossentropy'](sample_weight=sample_weight)
        else:
            return _LOSSES[self.loss](sample_weight=sample_weight)



#!/usr/bin/env python3

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import requests as req
import os
import sys
import ssl
import importlib.util
from pathlib import Path
import platform
import ibm_watson_machine_learning._wrappers.requests as requests
import json
import logging
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError

logger = logging.getLogger(__name__)

def process_heartbeat(client, **kwargs):
    try:
        config_dict = kwargs.get('config_dict')
        agg_info = config_dict.get('aggregator')
        wml_services_url = agg_info['ip'].split('/')[0]

        hearbeat_resp = requests.get("https://" + wml_services_url + "/wml_services/training/heartbeat")
        hearbeat_resp_json = json.loads(hearbeat_resp.content.decode("utf-8"))
        service_version = hearbeat_resp_json.get("service_implementation")

        if service_version.startswith("4.5") or service_version.startswith("45") or service_version.startswith("0.1.10000"):
            return "cpd450"
        if service_version.startswith("4.0"):
            version_details = service_version.split('-')
            if version_details[1] >= "202203":
                return "cpd408"
            if version_details[1] >= "202202":
                return "cpd407"
            if version_details[1] >= "202112":
                return "cpd406"
            elif version_details[1] >= "202110":
                return "cpd403"
            elif version_details[1] >= "202108":
                return "cpd402"
            else:
                return "cpd401"
        if service_version.startswith("35"):
            return "cpd35"

        return "cloud"

    except Exception as ex:
        logger.info("unable to process heartbeat")
        logger.exception(ex)
        return "cloud"


def import_diff(module_file_path):

    if "py38-tf24-pt17-sk23" in module_file_path:
        return
    pathlist = Path(module_file_path).rglob('*.py')
    for path in pathlist:
        # because path is object not string
        path_in_str = str(path)
        if not path_in_str.endswith("__init__.py"):
            module_name = ("ibmfl" + path_in_str.split("ibmfl")[2].replace(os.sep, "."))[:-3]
            module_spec = importlib.util.spec_from_file_location(
                module_name, path_in_str)
            loader = importlib.util.LazyLoader(module_spec.loader)
            module_spec.loader = loader
            module = importlib.util.module_from_spec(module_spec)
            sys.modules[module_name] = module
            module_spec.loader.exec_module(module)


def choose_software_version(software_spec):
    if software_spec == "runtime-22.1-py3.9" or software_spec == "12b83a17-24d8-5082-900f-0ab31fbfd3cb":
        return "py39-tf27-pt110-sk10"
    else:
        return "py38-tf24-pt17-sk23"


def check_python_version(platform_env):
    py_version = platform.python_version()
    if platform_env != "cloud" and platform_env != "cpd407" and platform_env != "cpd408" and platform_env != "cpd450":
        if not py_version.startswith("3.7.") and not py_version.startswith("3.8."):
            raise Exception("Python version {} is not supported for Federated Learning. "
                            "Parties must use a supported Python version to train the model. "
                            "For details, please refer to the documentation".format(py_version))
    else:
        if not py_version.startswith("3.7.") and not py_version.startswith("3.8.") and not py_version.startswith("3.9"):
            raise Exception("Python version {} is not supported for Federated Learning. "
                            "Parties must use a supported Python version to train the model. "
                            "For details, please refer to the documentation".format(py_version))


fl_path = os.path.abspath('.')
if fl_path not in sys.path:
    sys.path.append(fl_path)


class Party(WMLResource):
    """ 
    The Party class embodies a Federated Learning party, with methods to run, cancel, and query local training.
    Refer to the ``client.remote_training_system.create_party()`` API for more information about creating an
    instance of the Party class. 
    """

    SUPPORTED_PLATFORMS_MAP = {
        'cloud': "/cloud/ibmfl",
        'cpd35': "1.0.181",
        'cpd401': "1.0.122",
        'cpd402': "1.0.165",
        'cpd403': "/cpd403/ibmfl",
        'cpd406': "/cpd406/ibmfl",
        'cpd407': "/cpd406/ibmfl",
        'cpd408': "/cpd406/ibmfl",
        'cpd450': "/cpd406/ibmfl",
        'py38-tf24-pt17-sk23': "/py38-tf24-pt17-sk23/ibmfl", #default
        'py39-tf27-pt110-sk10': "/py39-tf27-pt110-sk10/ibmfl"
    }

    def __init__(self, client=None, **kwargs):

        libs_module = sys.modules['ibm_watson_machine_learning.libs']
        libs_location_list = libs_module.__path__

        # base location string, default to cloud location
        ibmfl_base_module_location = libs_location_list[0] + '/ibmfl'

        # get arg for platform
        calculated_platform_env = process_heartbeat(client=client, **kwargs)
        platform_env = kwargs.get('env', calculated_platform_env)
        print("Detected platform environment: {}".format(platform_env))

        #check python version
        check_python_version(platform_env)

        # process location
        ibmfl_module_location = ibmfl_base_module_location + self.SUPPORTED_PLATFORMS_MAP.get(platform_env)
        if not ibmfl_module_location.endswith("ibmfl"):
            raise Exception("Please use ibm-watson-machine-learning " + self.SUPPORTED_PLATFORMS_MAP.get(platform_env) )

        #check if using old connector script which is removed
        if not client:
            raise Exception("This version of the party connector script is outdated. "
                            "Please download the party connector script from your current Federated Learning experiment. "
                            "For more details, please refer to the documentation.")

        self.module_location = ibmfl_module_location
        self.args = kwargs
        self.Party = None
        self.connection = None
        self.log_level = None
        self.metrics_output = None

        if 'ibmfl' in sys.modules:
            del sys.modules['ibmfl']
        if 'ibmfl.party' in sys.modules:
            del sys.modules['ibmfl.party']
        if 'ibmfl.party.party' in sys.modules:
            del sys.modules['ibmfl.party.party']

        #install the general lib (which is the cloud version)
        module_name = 'ibmfl'
        module_spec = importlib.util.spec_from_file_location(
            module_name, ibmfl_base_module_location + '/py38-tf24-pt17-sk23/ibmfl/__init__.py')
        module = importlib.util.module_from_spec(module_spec)
        sys.modules[module_name] = module
        module_spec.loader.exec_module(module)

        WMLResource.__init__(self, __name__, client)
        self._client = client
        self.auth_token = "Bearer " + self._client.wml_token
        self.project_id = self._client.project_id
        self.log_level = kwargs.get("log_level", "ERROR")

    def start(self):
        self.Party.start()

    def run(self, aggregator_id=None, experiment_id=None, asynchronous=True, verify=True, timeout: int = 60 * 10):
        """
        Connect to a Federated Learning aggregator and run local training.
        Exactly one of aggregator_id and experiment_id must be supplied.  

        **Parameters**

        .. important::
           #. **aggregator_id**:  aggregator identifier\n
              * If aggregator_id is supplied, the party will connect to the given aggregator.\n
              **type**: str\n
           #. **experiment_id**:  experiment identifier\n
              * If experiment_id is supplied, the party will connect to the most recently created aggregator for the experiment.\n
              **type**: str\n
           #. **asynchronous**:\n
              * True  - party starts to run the job in the background and progress can be checked later.\n
              * False - method will wait until training is complete and then print the job status.\n
              **type**: bool\n
           #. **verify**:  verify certificate\n
              **type**: bool\n
           #. **timeout**:  timeout in seconds.\n
              * If the aggregator is not ready within timeout seconds, exit.\n
              **type**: int, or None for no timeout\n


        **Examples**
         >>> party.run( aggregator_id = "69500105-9fd2-4326-ad27-7231aeb37ac8", asynchronous = True, verify = True )
         >>> party.run( experiment_id = "2466fa06-1110-4169-a166-01959adec995", asynchronous = False )

        """
        from ibmfl.exceptions import FLException
        from datetime import datetime
        import time

        timeout_time = None if timeout is None else timeout + time.time()
        if ( experiment_id is None and aggregator_id is None ) or ( experiment_id is not None and aggregator_id is not None ) :
            raise FLException("Exactly one of aggregator_id and experiment_id must be supplied")

        if ( experiment_id is not None ):
            while True: 
                try:
                    details = self._client.training.get_details(get_all=True,training_definition_id=experiment_id, _internal=True)['resources']
                    details = [ d for d in details if d['entity']['status']['state'] in ['accepting_parties','pending','running'] ]
                    if not details :
                        if timeout_time and timeout_time < time.time() :
                            raise FLException("Cannot find an aggregator for experiment %s" % experiment_id)
                        else :
                            logger.info("Cannot find an aggregator for experiment %s . Retrying." % experiment_id)
                            time.sleep(30)
                            continue
                    else :
                        aggregator_id = max( [(t['metadata']['id'],datetime.strptime(t['metadata']['created_at'],'%Y-%m-%dT%H:%M:%S.%fZ'))
                                          for t in details],key=lambda d:d[1])[0]
                        logger.info("Using aggregator id %s" % aggregator_id) 
                        break
                except Exception as ex:
                    logger.exception(ex)
                    raise FLException(str(ex))


        #import the changed files in the desired lib version

        config_dict = self.args.get('config_dict')
        metrics_config = {
            "name": "WMLMetricsRecorder",
            "path": "ibmfl.party.metrics.metrics_recorder",
            "output_file": self.metrics_output,
            "output_type": "json",
            'compute_pre_train_eval': False,
            'compute_post_train_eval': False
        }
        if 'metrics_recorder' not in config_dict:
            config_dict["metrics_recorder"] = metrics_config
        wml_services_url = config_dict.get('aggregator').get('ip').split('/')[0]
        agg_info = wml_services_url + "/ml/v4/trainings/" + aggregator_id
        config_dict['aggregator']['ip'] = agg_info
        self.args['config_dict'] = config_dict

        #verify ssl context
        if verify:
            try:
                req.get("https://" + wml_services_url + "/wml_services/training/heartbeat", verify=verify)
            except requests.exceptions.SSLError as ex:
                logger.error(str(ex))
                raise FLException("No valid certificate detected. Please replace the default certificate with your own "
                                  "TLS certificate, or set verify to False at your own risk. For more details, please see "
                                  "https://www.ibm.com/docs/en/cloud-paks/cp-data/4.0?topic=client-using-custom-tls-certificate-connect-platform")

        #check for aggregator state and start job
        try:
            training_status = self._client.training.get_status(aggregator_id)
            state = training_status['state']
            ready = False
            if state == "pending":
                while state == "pending" and ( not timeout_time or timeout_time > time.time() ) :
                    logger.info("Waiting for aggregator accepting parties state..")
                    time.sleep(10)
                    training_status = self._client.training.get_status(aggregator_id)
                    state = training_status['state']
                if state != "accepting_parties":
                    raise FLException("The current state of training %s is %s, so the party is not able to start a job." % (aggregator_id,state))
                ready = True
            elif state == "running" or state == "accepting_parties":
                ready = True
            else:
                raise FLException("The current state of training %s is %s, so the party is not able to start a job." % (aggregator_id,state))

            if ready:
                if "cpd406" in self.module_location or "cloud" in self.module_location:
                    details = self._client.training.get_details(aggregator_id, _internal=True)
                    fl_entity = details["entity"]["federated_learning"]
                    if "software_spec" in fl_entity:
                        software_spec = fl_entity["software_spec"]["name"] if "name" in fl_entity["software_spec"] else fl_entity["software_spec"]["id"]
                    else:
                        software_spec = "runtime-22.1-py3.9"
                    platform_env = choose_software_version(software_spec)
                    print("Loading {} environment..".format(platform_env))
                    self.module_location = '/'.join(self.module_location.split('/')[0:-2]) + self.SUPPORTED_PLATFORMS_MAP.get(platform_env)
                import_diff(self.module_location)
                from ibmfl.party.party import Party
                self.Party = Party(**self.args, token=self.auth_token, self_signed_cert=not verify, log_level=self.log_level)
                self.connection = self.Party.connection
                self.start()
            else:
                raise FLException("The current state of training %s is %s, so the party is not able to start a job." % (aggregator_id,state))
            #wait for the job to finish if synchrounous
            if not asynchronous:
                while "completed" != state and "failed" != state and "canceled" != state and self.is_running():
                    training_status = self._client.training.get_status(aggregator_id)
                    state = training_status['state']
                    time.sleep(10)
                logger.info("The training finishes with %s status" % state)
        except FLException as ex:
            raise FLException(str(ex))
        except Exception as ex:
            logger.info("The party failed to start training")
            logger.exception(ex)

    def monitor_logs(self, log_level="INFO"):
        """
        Enable logging of the training job to standard output.
        This method should be called before calling the run() method.

        **Parameters**

        .. important::
           #. **log_level**:  log level specified by user\n
              **type**: str\n

        **Output**

        .. important::
            **returns**: This method only outputs the logs to stdout\n
            **return type**: None\n

        **Example**
         >>> party.monitor_logs()

        """
        self.log_level = log_level

    def monitor_metrics(self, metrics_file="-"):
        """
        Enable output of training metrics.

        **Parameters**

        .. important::
           #. **metrics_file**:  a filename specified by user to which the metrics should be written\n
              **type**: str\n

        **Output**

        .. important::
            **returns**: This method outputs the metrics to stdout if a filename is not specified\n
            **return type**: None\n

        **Example**
         >>> party.monitor_metrics()

        """
        self.metrics_output = metrics_file

    def is_running(self):
        """
        Check if the training job is running

        **Output**

        .. important::
             **returns**: If the job is running\n
             **return type**: bool\n

        **Example**
         >>> party.is_running()

        """
        return not self.connection.stopped

    def get_round(self):  

        """
        Get the current round number

        **Output**

        .. important::
             **returns**: The current round number\n
             **return type**: int\n

        **Example**
         >>> party.get_round()

        """
        return self.Party.proto_handler.metrics_recorder.get_round_no()

    def cancel(self):
        """
        Stop the local connection to the training on the party side

        **Example**
         >>> party.cancel()

        """
        self.Party.stop_connection()



# coding: utf-8

"""
    IBM Watson Machine Learning REST API

    ## Authorization  ### IBM Watson Machine Learning Credentials (ML Credentials)  The IBM Watson Machine Learning Credentials are available for the Bluemix user for each bound application or requested service key. You will find them in the VCAP information as well you can access them using Cloud Foundry API.  Here is the example of the ML Credentials:  ```json {   \"url\": \"https://ibm-watson-ml.mybluemix.net\",   \"username\": \"c1ef4b80-2ee2-458e-ab92-e9ca97ec657d\",   \"password\": \"edb699da-8595-406e-bae0-74a834fa4d34\",    \"access_key\": \"0uUQPsbQozcci4uwRI7xo0jQnSNOM9YSk....\" } ```  - `url` - the base WML API url - `username` / `password` - the service credentials required to generate the token - `access_key` - the access key used by previous version of the service API (ignored)  The `username` / `password` pair are used to access the Token Endpoint (using HTTP Basic Auth) and obtain the service token (see below). Example:  `curl --basic --user c1ef4b80-2ee2-458e-ab92-e9ca97ec657d:edb699da-8595-406e-bae0-74a834fa4d34 https://ibm-watson-ml.mybluemix.net/v2/identity/token`  ### IBM Watson Machine Learning Token (ML Token)  The IBM Watson Machine Learning REST API is authorized with ML token obtained from the Token Endpoint. The ML token is used as a baerer token send in `authorization` header.  Use WML service credentials (username, password) to gather the token from:   `/v2/identity/token` (see example above).  The token is self-describing JWT (JSON Web Tokens) protected by digital signature for authentication. It holds information required for a service tenant identification. Each ML micro-service is able to verify the token with the public key without request to the token endpoint and firing a database query. The ML service token (ML token) contains the expiration time what simplifies implementation of the access revocation.  ## Spark Instance  The IBM Watson ML co-operates with the Spark as a Service to make calculation and deploy pipeline models. Each API method that requires the Spark service instance accepts a custom header: `X-Spark-Service-Instance` where the Spark instance data like credentials, kernel ID and version can be specified. The header value is a base64 encoded string with the JSON data in the following format:    ```   {     \"credentials\": {       \"tenant_id\": \"sf2c-xxxxx-05b1d10fb12b\",       \"cluster_master_url\": \"https://spark.stage1.bluemix.net\",       \"tenant_id_full\": \"xxxxx-a94d-4f20-bf2c-xxxxxx-xxxx-4c65-a156-05b1d10fb12b\",       \"tenant_secret\": \"xxxx-86fd-40cd-xxx-969aafaeb505\",       \"instance_id\": \"xxx-a94d-xxx-bf2c-xxxx\",       \"plan\": \"ibm.SparkService.PayGoPersonal\"     },     \"version\": \"2.0\",     \"kernelId\": \"xxx-a94d-xxx-bf2c-xxxx\"   }   ```  The fields are: - `credentials` - from the VCAP information of the Spark service instance - `version` - requested Spark version (possible values: 2.0) - `kernelId` (optional) - the Spark kernel ID is only required by actions that operates on running Spark kernel.   This field is redundant when creating any kind of deployment. 

    OpenAPI spec version: 2.0.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pprint import pformat
from six import iteritems
import re


class ModelOutputEntity(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, name=None, description=None, author=None, type=None, runtime_environment=None, training_data_schema=None, label_col=None, input_data_schema=None, pipeline_version=None, versions_href=None, latest_version=None, output_data_schema=None):
        """
        ModelOutputEntity - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'name': 'str',
            'description': 'str',
            'author': 'ArtifactAuthor',
            'type': 'str',
            'runtime_environment': 'str',
            'training_data_schema': 'object',
            'label_col': 'str',
            'input_data_schema': 'object',
            'pipeline_version': 'ModelOutputEntityPipelineVersion',
            'versions_href': 'str',
            'latest_version': 'ArtifactVersionMetadata',
            'output_data_schema' :'object'
        }

        self.attribute_map = {
            'name': 'name',
            'description': 'description',
            'author': 'author',
            'type': 'type',
            'runtime_environment': 'runtimeEnvironment',
            'training_data_schema': 'trainingDataSchema',
            'label_col': 'labelCol',
            'input_data_schema': 'inputDataSchema',
            'pipeline_version': 'pipelineVersion',
            'versions_href': 'versionsHref',
            'latest_version': 'latestVersion',
            'output_data_schema' : 'outputDataSchema'
        }

        self._name = name
        self._description = description
        self._author = author
        self._type = type
        self._runtime_environment = runtime_environment
        self._training_data_schema = training_data_schema
        self._label_col = label_col
        self._input_data_schema = input_data_schema
        self._pipeline_version = pipeline_version
        self._versions_href = versions_href
        self._latest_version = latest_version
        self._output_data_schema = output_data_schema

    @property
    def name(self):
        """
        Gets the name of this ModelOutputEntity.
        Name of the model

        :return: The name of this ModelOutputEntity.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this ModelOutputEntity.
        Name of the model

        :param name: The name of this ModelOutputEntity.
        :type: str
        """

        self._name = name

    @property
    def description(self):
        """
        Gets the description of this ModelOutputEntity.
        Description of the model

        :return: The description of this ModelOutputEntity.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this ModelOutputEntity.
        Description of the model

        :param description: The description of this ModelOutputEntity.
        :type: str
        """

        self._description = description

    @property
    def author(self):
        """
        Gets the author of this ModelOutputEntity.
        Author of the pipeline

        :return: The author of this ModelOutputEntity.
        :rtype: ArtifactAuthor
        """
        return self._author

    @author.setter
    def author(self, author):
        """
        Sets the author of this ModelOutputEntity.
        Author of the pipeline

        :param author: The author of this ModelOutputEntity.
        :type: ArtifactAuthor
        """

        self._author = author

    @property
    def type(self):
        """
        Gets the type of this ModelOutputEntity.
        Type of the ML pipeline model

        :return: The type of this ModelOutputEntity.
        :rtype: ModelType
        """
        return self._type

    @type.setter
    def type(self, type):
        """
        Sets the type of this ModelOutputEntity.
        Type of the ML pipeline model

        :param type: The type of this ModelOutputEntity.
        :type: ModelType
        """

        self._type = type

    @property
    def runtime_environment(self):
        """
        Gets the runtime_environment of this ModelOutputEntity.
        runtime environment of caller

        :return: The runtime_environment of this ModelOutputEntity.
        :rtype: RuntimeEnvironment
        """
        return self._runtime_environment

    @runtime_environment.setter
    def runtime_environment(self, runtime_environment):
        """
        Sets the runtime_environment of this ModelOutputEntity.
        runtime environment of caller

        :param runtime_environment: The runtime_environment of this ModelOutputEntity.
        :type: RuntimeEnvironment
        """

        self._runtime_environment = runtime_environment

    @property
    def training_data_schema(self):
        """
        Gets the training_data_schema of this ModelOutputEntity.
        JSON schema of the training data

        :return: The training_data_schema of this ModelOutputEntity.
        :rtype: TrainingDataSchema
        """
        return self._training_data_schema

    @training_data_schema.setter
    def training_data_schema(self, training_data_schema):
        """
        Sets the training_data_schema of this ModelOutputEntity.
        JSON schema of the training data

        :param training_data_schema: The training_data_schema of this ModelOutputEntity.
        :type: TrainingDataSchema
        """

        self._training_data_schema = training_data_schema

    @property
    def label_col(self):
        """
        Gets the label_col of this ModelOutputEntity.
        Name of the label column

        :return: The label_col of this ModelOutputEntity.
        :rtype: str
        """
        return self._label_col

    @label_col.setter
    def label_col(self, label_col):
        """
        Sets the label_col of this ModelOutputEntity.
        Name of the label column

        :param label_col: The label_col of this ModelOutputEntity.
        :type: str
        """

        self._label_col = label_col

    @property
    def input_data_schema(self):
        """
        Gets the input_data_schema of this ModelOutputEntity.
        JSON schema of the input data (if not provided the training data schema is used)

        :return: The input_data_schema of this ModelOutputEntity.
        :rtype: InputDataSchema
        """
        return self._input_data_schema

    @input_data_schema.setter
    def input_data_schema(self, input_data_schema):
        """
        Sets the input_data_schema of this ModelOutputEntity.
        JSON schema of the input data (if not provided the training data schema is used)

        :param input_data_schema: The input_data_schema of this ModelOutputEntity.
        :type: InputDataSchema
        """

        self._input_data_schema = input_data_schema

    @property
    def pipeline_version(self):
        """
        Gets the pipeline_version of this ModelOutputEntity.


        :return: The pipeline_version of this ModelOutputEntity.
        :rtype: ModelOutputEntityPipelineVersion
        """
        return self._pipeline_version

    @pipeline_version.setter
    def pipeline_version(self, pipeline_version):
        """
        Sets the pipeline_version of this ModelOutputEntity.


        :param pipeline_version: The pipeline_version of this ModelOutputEntity.
        :type: ModelOutputEntityPipelineVersion
        """

        self._pipeline_version = pipeline_version

    @property
    def versions_href(self):
        """
        Gets the versions_href of this ModelOutputEntity.
        Href to the model versions collection

        :return: The versions_href of this ModelOutputEntity.
        :rtype: str
        """
        return self._versions_href

    @versions_href.setter
    def versions_href(self, versions_href):
        """
        Sets the versions_href of this ModelOutputEntity.
        Href to the model versions collection

        :param versions_href: The versions_href of this ModelOutputEntity.
        :type: str
        """

        self._versions_href = versions_href

    @property
    def latest_version(self):
        """
        Gets the latest_version of this ModelOutputEntity.
        the latest artifact version metadata

        :return: The latest_version of this ModelOutputEntity.
        :rtype: ArtifactVersionMetadata
        """
        return self._latest_version

    @latest_version.setter
    def latest_version(self, latest_version):
        """
        Sets the latest_version of this ModelOutputEntity.
        the latest artifact version metadata

        :param latest_version: The latest_version of this ModelOutputEntity.
        :type: ArtifactVersionMetadata
        """

        self._latest_version = latest_version

    @property
    def output_data_schema(self):
        """
        Gets the output_data_schema of this ModelOutputEntity.
        JSON schema of the output data

        :return: The output_data_schema of this ModelOutputEntity.
        :rtype: OutputDataSchema
        """
        return self._output_data_schema

    @output_data_schema.setter
    def output_data_schema(self, output_data_schema):
        """
        Sets the output_data_schema of this ModelOutputEntity.
        JSON schema of the output data

        :param output_data_schema: The output_data_schema of this ModelOutputEntity.
        :type: OutputDataSchema
        """

        self._output_data_schema = output_data_schema


    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  (C) Copyright IBM Corp. 2022.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
import matplotlib.pyplot as plt
from IPython.display import display, clear_output


def calculate_cum_limits(scores):
    def lower_limit(scores):
        mean = np.mean(scores)
        std = np.std(scores)
        return mean - std

    def upper_limit(scores):
        mean = np.mean(scores)
        std = np.std(scores)
        return mean + std

    n = len(scores) + 1
    lower_limits = [lower_limit(scores[:i]) for i in range(1, n)]
    upper_limits = [upper_limit(scores[:i]) for i in range(1, n)]

    return lower_limits, upper_limits


def plot_subcurve(fig, ax, values, x=None, color='b', title='Learning Curve', xlabel='Training batches', ylabel='Score',
                  refresh=True):
    if x is None:
        x = list(range(0, len(values)))

    lower_limits, upper_limits = calculate_cum_limits(values)
    ax.cla()
    ax.grid()
    ax.plot(x, values, "o-")
    ax.fill_between(
        x,
        lower_limits,
        upper_limits,
        alpha=0.1,
        color=color,
    )
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    if refresh:
        display(fig)
        clear_output(wait=True)


def plot_learning_curve(fig, axes, scores, fit_times=None):
    if isinstance(axes, np.ndarray) and fit_times is not None:
        plot_subcurve(fig, axes[0], scores)
        if len(fit_times) < len(scores):
            x = list(range(len(scores) - len(fit_times), len(scores)))
            plot_subcurve(fig, axes[1], fit_times, x, color='g', title='Scalability of the model',
                          xlabel='Training batches', ylabel='Fit time')
        else:
            plot_subcurve(fig, axes[1], fit_times, color='g', title='Scalability of the model', xlabel='Training batches',
                          ylabel='Fit time')
    else:
        plot_subcurve(fig, axes, scores)



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import shutil
import zipfile

from ibm_watson_machine_learning.utils.autoai.utils import try_import_joblib
from ibm_watson_machine_learning.libs.repo.util.unique_id_gen import uid_generate


class HybridArtifactLoader(object):
    def load(self, artifact_queryparam):

        if artifact_queryparam is None:
           return self.extract_content(artifact_queryparam, HybridArtifactLoader._load_content)

        if artifact_queryparam is not None and artifact_queryparam == "full":
            return self.extract_content(artifact_queryparam, HybridArtifactLoader._load_content)

        if artifact_queryparam is not None and artifact_queryparam == "pipeline_model":
            return self.extract_content_json()

    def extract_content(self, queryparam_val, callback):
        directory_name = 'artifact'
        try:
            shutil.rmtree(directory_name)
        except:
            pass

        try:
            id_length = 20
            dir_id = uid_generate(id_length)
            model_dir_name = directory_name + dir_id

            gz_file_name = '{}/artifact_content.tar.gz'.format(model_dir_name)

            input_stream = None
            os.makedirs(model_dir_name)
            if queryparam_val is None:
                input_stream = self.reader().read()
            if queryparam_val is not None and queryparam_val == 'full':
                input_stream = self._content_reader_gzip()
            file_content = input_stream.read()
            gz_f = open(gz_file_name, 'wb+')
            gz_f.write(file_content)
            gz_f.close()
            if queryparam_val is None:
                self.reader().close()

            with zipfile.ZipFile(gz_file_name) as zip_ref:
                zip_ref.extractall(model_dir_name)
            artifact_instance = callback(model_dir_name)

            shutil.rmtree(model_dir_name)
            return artifact_instance
        except Exception as ex:
            shutil.rmtree(model_dir_name)
            raise ex

    def extract_content_json(self):
        directory_name = 'artifact'
        try:
            shutil.rmtree(directory_name)
        except:
            pass

        try:
            id_length = 20
            dir_id = uid_generate(id_length)
            model_dir_name = directory_name + dir_id

            output_file_name = '{}/pipeline_model.json'.format(model_dir_name)

            os.makedirs(model_dir_name)
            input_stream = self._content_reader_json()
            file_content = input_stream.read()
            gz_f = open(output_file_name, 'wb+')
            gz_f.write(file_content)
            gz_f.close()
            return output_file_name
        except Exception as ex:
            shutil.rmtree(model_dir_name)
            raise ex

    def _content_reader_gzip(self):
        if self._download_href is not None:
            if self._download_href.__contains__("models"):
                download_url = self._download_href + f"&space_id={self.meta.prop('space_id')}"\
                                                     "&content_format=pipeline-node&pipeline_node_id=automl"
                return self.client.repository_api.download_artifact_content_v4_cloud(download_url, 'true')

    def _content_reader_json(self):
        if self._download_href is not None:
            if self._download_href.__contains__("models"):
                download_url = self._download_href + f"&space_id={self.meta.prop('space_id')}&content_format=native"
                return self.client.repository_api.download_artifact_content_v4_cloud(download_url, 'true')

    @staticmethod
    def _load_content(content_dir):
        """
        Load AutoAI pipeline into a local runtime.
        """
        joblib = try_import_joblib()
        extracted_file = [file for file in os.listdir(content_dir) if file.endswith(".pickle")][0]
        return joblib.load(os.path.join(content_dir, extracted_file))



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pprint import pformat
from six import iteritems

class ContentStatus(object):
    def __init__(self, state=None, message=None, failure=None):
        """
        ModelMetrics - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'state': 'str',
            'message': 'str',
            'failure': 'ErrorSchemaRepository'
        }

        self.attribute_map = {
            'state': 'state',
            'message': 'message',
            'failure': 'failure'
        }

        self._state = state
        self._message = message
        self._failure = failure

    @property
    def state(self):
        """
        Gets the code of this ContentStatus.


        :return: The code of this ContentStatus.
        :rtype: str
        """
        return self._state

    @state.setter
    def state(self, state):
        """
        Sets the state of this ContentStatus.


        :param state: The state of this ContentStatus.
        :type: str
        """
        self._state = state

    @property
    def message(self):
        """
        Gets the error message of this ContentStatus.


        :return: The error message of this ContentStatus.
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """
        Sets the message of this ContentStatus.


        :param message: The message of this ContentStatus.
        :type: str
        """
        self._message = message

    @property
    def failure(self):
        """
        Gets the details for the failure of this ContentStatus.


        :return: The the details for the failure of this ContetnStatus.
        :rtype: ErrorSchemaRepository
        """
        return self._failure

    @failure.setter
    def failure(self, failure):
        """
        Sets the failure details of this ContentStatus.


        :param failure: The failure details of this ContentStatus.
        :type: ErrorSchemaRepository
        """
        self._failure = failure

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other